--  Copyright (C) 2002-2003 David Roundy
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2, or (at your option)
--  any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; see the file COPYING.  If not, write to
--  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
--  Boston, MA 02110-1301, USA.

{-# LANGUAGE CPP, PatternGuards #-}

module Darcs.UI.Commands.Record ( record, commit, getDate, getLog,
                               askAboutDepends
                             ) where

import Prelude hiding ( catch )

import Control.Applicative ( (<$>) )
import Control.Exception ( handleJust, catch, IOException )
import Control.Monad ( when, unless, void )
import System.IO ( stdin )
import Data.List ( sort, isPrefixOf, union )
import Data.Char ( ord )
import System.Exit ( exitFailure, exitSuccess, ExitCode(..) )
import System.Directory ( removeFile )
import Data.Maybe ( isJust, catMaybes )
import qualified Data.ByteString as B ( hPut )

import Darcs.Repository.Lock
    ( readLocaleFile
    , writeLocaleFile
    , worldReadableTemp
    , appendToFile
    )
import Darcs.Patch.PatchInfoAnd ( info, n2pia, PatchInfoAnd )
import Darcs.Repository
    ( Repository
    , withRepoLock
    , RepoJob(..)
    , readTentativeRepo
    , tentativelyAddPatch
    , finalizeRepositoryChanges
    , invalidateIndex
    , unrecordedChangesWithPatches
    , readRecorded
    , listRegisteredFiles
    )
import Darcs.Patch ( RepoPatch, Patchy, PrimOf, PrimPatch, namepatch, summaryFL, anonymous,
                     adddeps, fromPrims )
import Darcs.Patch.Set ( PatchSet(..) )
import Darcs.Patch.Witnesses.Sealed
import Darcs.Patch.Witnesses.Eq ( unsafeCompare )
import Darcs.Patch.Witnesses.Ordered
    ( FL(..), (:>)(..), (+>+),
    reverseRL, mapFL, mapFL_FL, nullFL )
import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP )
import Darcs.Patch.Info ( PatchInfo )
import Darcs.Patch.Apply( ApplyState )
import Darcs.Patch.Split ( primSplitter )
import Darcs.Patch.Choices ( patchChoicesTps, tpPatch,
                             forceFirsts, getChoices, tag )
import Darcs.UI.External ( editFile )
import Darcs.UI.SelectChanges
    ( selectChanges
    , WhichChanges(..)
    , selectionContext
    , selectionContextPrim
    , runSelection
    )
import Darcs.Util.Path ( FilePathLike, SubPath, toFilePath, AbsolutePath )
import Darcs.UI.Commands
    ( DarcsCommand(..)
    , nodefaults
    , commandStub
    , setEnvDarcsFiles
    , setEnvDarcsPatches
    , amInHashedRepository
    )
import Darcs.UI.Commands.Util ( announceFiles, filterExistingFiles,
                                testTentativeAndMaybeExit )
import Darcs.UI.Arguments
    ( DarcsFlag( PromptLongComment
               , NoEditLongComment
               , EditLongComment
               , LogFile
               , Pipe
               , Quiet
               , PatchName
               , AskDeps
               , All
               , LookForMoves
               , LookForReplaces
               )
    , fileHelpAuthor
    , getAuthor
    , workingRepoDir
    , lookforadds
    , lookformoves
    , lookforreplaces
    , fixSubPaths
    , askLongComment
    , askdeps
    , patchSelectFlag
    , allPipeInteractive
    , test
    , author
    , patchnameOption
    , umaskOption
    , ignoretimes
    , nocompress
    , rmlogfile
    , logfile
    , setScriptsExecutableOption
    , unified
    , diffAlgorithm
    )
import qualified Darcs.UI.Arguments as A ( leaveTestDir )
import Darcs.UI.Flags
    (willRemoveLogFile, diffingOpts, compression, verbosity, umask,
     useCache, lookForAdds )
import Darcs.Repository.Flags ( UpdateWorking (..), DryRun(NoDryRun)  )
import Darcs.Repository.Util ( getMovesPs, getReplaces )
import Darcs.Util.Exception ( clarifyErrors )
import Darcs.Util.Prompt ( askUser, promptYorn )
import Darcs.Util.Progress ( debugMessage)
import Darcs.Patch.Progress ( progressFL )
import Darcs.Util.Printer ( hPutDocLn, text, ($$), prefixLines )
import Darcs.Util.ByteString ( encodeLocale )
import Darcs.Util.IsoDate ( getIsoDateTime, cleanLocalDate )
import qualified Darcs.Util.Ratified as Ratified ( hGetContents )
import Storage.Hashed.Tree( Tree )

#include "impossible.h"

recordDescription :: String
recordDescription = "Create a patch from unrecorded changes."

recordHelp :: String
recordHelp =
 "The `darcs record` command is used to create a patch from changes in\n" ++
 "the working tree.  If you specify a set of files and directories,\n" ++
 "changes to other files will be skipped.\n" ++
 "\n" ++ recordHelp' ++
 "\n" ++ recordHelp''

record :: DarcsCommand
record = DarcsCommand {commandProgramName = "darcs",
                       commandName = "record",
                       commandHelp = recordHelp,
                       commandDescription = recordDescription,
                       commandExtraArgs = -1,
                       commandExtraArgHelp = ["[FILE or DIRECTORY]..."],
                       commandCommand = recordCmd,
                       commandPrereq = amInHashedRepository,
                       commandGetArgPossibilities = listRegisteredFiles,
                       commandArgdefaults = nodefaults,
                       commandAdvancedOptions = [logfile, rmlogfile,
                                                   nocompress, ignoretimes,
                                                   umaskOption,
                                                   setScriptsExecutableOption],
                       commandBasicOptions = [patchnameOption, author,
                                               test,
                                               A.leaveTestDir,
                                               allPipeInteractive,
                                               askdeps,
                                               askLongComment,
                                               lookforadds,
                                               lookformoves,lookforreplaces,
                                               workingRepoDir, unified,
                                               diffAlgorithm]}

commitDescription :: String
commitDescription = "Redirect the user to record, push or send."

commitHelp :: String
commitHelp =
 "This command does not do anything.\n"++
 "If you want to save changes locally, use the `darcs record` command.\n"++
 "If you want to save a recorded patch to another repository, use the\n"++
 "`darcs push` or `darcs send` commands instead.\n"

commit :: DarcsCommand
commit = commandStub "commit" commitHelp commitDescription record

recordCmd :: (AbsolutePath, AbsolutePath) -> [DarcsFlag] -> [String] -> IO ()
recordCmd fps opts args = do
    checkNameIsNotOption opts
    withRepoLock NoDryRun (useCache opts) YesUpdateWorking (umask opts) $ RepoJob $ \(repository :: Repository p wR wU wR) -> do
    files <- if null args then return Nothing
        else Just . sort <$> fixSubPaths fps args
    when (files == Just []) $ fail "No valid arguments were given."
    announceFiles files "Recording changes in"
    existing_files <- maybe (return Nothing)
        (fmap Just . filterExistingFiles repository (lookForAdds opts)) files
    when (existing_files == Just []) $
       fail "None of the files you specified exist!"
    debugMessage "About to get the unrecorded changes."
    Sealed replacePs <- if LookForReplaces `elem` opts
      then getReplaces (diffingOpts opts) repository files
      else return (Sealed NilFL)
    movesPs <- if LookForMoves `elem` opts
        then getMovesPs repository files
        else return NilFL
    changes <- unrecordedChangesWithPatches (diffingOpts opts) repository files
                 movesPs (unsafeCoerceP replacePs :: FL (PrimOf p) wR wR)
    debugMessage "I've gotten unrecorded."
    case allow_empty_with_askdeps changes of
      Nothing -> do when (Pipe `elem` opts) $
                        -- We need to grab any input waiting for us, since we
                        -- might break scripts expecting to send it to us; we
                        -- don't care what that input is, though.
                        void (getDate opts)
                    putStrLn "No changes!"
                    exitFailure
      Just ch -> doRecord repository opts existing_files ch
    where allow_empty_with_askdeps :: FL p wX wY -> Maybe (FL p wX wY)
          allow_empty_with_askdeps NilFL
              | AskDeps `elem` opts = Just NilFL
              | otherwise = Nothing
          allow_empty_with_askdeps p = Just p

 -- check that what we treat as the patch name is not accidentally a command
 -- line flag
checkNameIsNotOption :: [DarcsFlag] -> IO ()
checkNameIsNotOption opts = do
    let patchNames = [n | PatchName n <- opts]
    when (length patchNames == 1) $ do
        let n = head patchNames
            oneLetterName = length n == 1 || (length n == 2 && head n == '-')
        when (oneLetterName && (All `notElem` opts)) $ do
            confirmed <- promptYorn $ "You specified " ++ show n ++ " as the patch name. Is that really what you want?"
            unless confirmed $ putStrLn "Okay, aborting the record." >> exitFailure


doRecord :: (RepoPatch p, ApplyState p ~ Tree, ApplyState (PrimOf p) ~ Tree)
         => Repository p wR wU wR -> [DarcsFlag] -> Maybe [SubPath] -> FL (PrimOf p) wR wX -> IO ()
doRecord repository opts files ps = do
    let make_log = worldReadableTemp "darcs-record"
    date <- getDate opts
    my_author <- getAuthor opts
    debugMessage "I'm slurping the repository."
    debugMessage "About to select changes..."
    pristine <- readRecorded repository
    (chs :> _ ) <- runSelection (selectChanges First ps) $
                  selectionContextPrim "record" opts (Just primSplitter)
                                       (map toFilePath <$> files)
                                       (Just pristine)
    when (is_empty_but_not_askdeps chs) $
              do putStrLn "Ok, if you don't want to record anything, that's fine!"
                 exitSuccess
    handleJust onlySuccessfulExits (\_ -> return ()) $
             do deps <- if AskDeps `elem` opts
                        then askAboutDepends repository chs opts []
                        else return []
                when (AskDeps `elem` opts) $ debugMessage "I've asked about dependencies."
                if nullFL chs && null deps
                  then putStrLn "Ok, if you don't want to record anything, that's fine!"
                  else do setEnvDarcsFiles chs
                          (name, my_log, logf) <- getLog opts Nothing make_log chs
                          debugMessage ("Patch name as received from getLog: " ++ show (map ord name))
                          doActualRecord repository opts name date
                                 my_author my_log logf deps chs
    where is_empty_but_not_askdeps l
              | AskDeps `elem` opts = False
                                      -- a "partial tag" patch; see below.
              | otherwise = nullFL l

doActualRecord :: (RepoPatch p, ApplyState p ~ Tree)
               => Repository p wR wU wR -> [DarcsFlag] -> String -> String -> String
               -> [String] -> Maybe String
               -> [PatchInfo] -> FL (PrimOf p) wR wX -> IO ()
doActualRecord repository opts name date my_author my_log logf deps chs =
              do debugMessage "Writing the patch file..."
                 mypatch <- namepatch date name my_author my_log $
                            fromPrims $ progressFL "Writing changes:" chs
                 let pia = n2pia $ adddeps mypatch deps
                 -- We don't care about the returned updated repository
                 _ <- tentativelyAddPatch repository (compression opts) (verbosity opts) YesUpdateWorking
                           $ pia
                 invalidateIndex repository
                 debugMessage "Applying to pristine..."
                 testTentativeAndMaybeExit repository opts
                      ("you have a bad patch: '" ++ name ++ "'")
                      "record it" (Just failuremessage)
                 finalizeRepositoryChanges repository YesUpdateWorking (compression opts)
                                    `clarifyErrors` failuremessage
                 debugMessage "Syncing timestamps..."
                 when (isJust logf) $ removeFile (fromJust logf)
                 when (Quiet `notElem` opts) $
                     putStrLn $ "Finished recording patch '"++name++"'"
                 setEnvDarcsPatches (pia :>: NilFL)
    where failuremessage = "Failed to record patch '"++name++"'" ++
                           case logf of Just lf -> "\nLogfile left in "++lf++"."
                                        Nothing -> ""

recordHelp' :: String
recordHelp' = unlines
 [ "Every patch has a name, an optional description, an author and a date."
 , ""
 , "Darcs will launch a text editor (see `darcs help environment`) after the"
 , "interactive selection, to let you enter the patch name (first line) and"
 , "the patch description (subsequent lines)."
 , ""
 , "The patch name should be a short sentence that concisely describes the"
 , "patch, such as \"Add error handling to main event loop.\"  You can"
 , "supply it in advance with the `-m` option, in which case no text editor"
 , "is launched, unless you use the `--edit-long-comment` option."
 , ""
 , "The patch description is an optional block of free-form text.  It is"
 , "used to supply additional information that doesn't fit in the patch"
 , "name.  For example, it might include a rationale of WHY the change was"
 , "necessary."
 , ""
 , "A technical difference between patch name and patch description, is"
 , "that matching with the flag `-p` is only done on patch names."
 , ""
 , "Finally, the `--logfile` option allows you to supply a file that"
 , "already contains the patch name and patch description.  This is"
 , "useful if a previous record failed and left a `darcs-record-0` file."
 , ""
 , unlines fileHelpAuthor
 , ""
 , "If you want to manually define any dependencies for your patch, you can"
 , "use the `--ask-deps` flag, and darcs will ask you for the patch's"
 , "dependencies.  A patch with specific dependencies can be empty."
 , ""
 , "The patch date is generated automatically.  It can only be spoofed by"
 , "using the `--pipe` option."
 , ""
 , "If you run record with the `--pipe` option, you will be prompted for"
 , "the patch date, author, and the long comment. The long comment will extend"
 , "until the end of file or stdin is reached (ctrl-D on Unixy systems, ctrl-Z"
 , "on systems running a Microsoft OS)."
 , ""
 , "This interface is intended for scripting darcs, in particular for writing"
 , "repository conversion scripts.  The prompts are intended mostly as a useful"
 , "guide (since scripts won't need them), to help you understand the format in"
 , "which to provide the input. Here's an example of what the `--pipe`"
 , "prompts look like:"
 , ""
 , "    What is the date? Mon Nov 15 13:38:01 EST 2004"
 , "    Who is the author? David Roundy"
 , "    What is the log? One or more comment lines"
 ]

getDate :: [DarcsFlag] -> IO String
getDate opts
 | Pipe `elem` opts = cleanLocalDate =<< askUser "What is the date? "
getDate _ = getIsoDateTime

data PName = FlagPatchName String | PriorPatchName String | NoPatchName


-- | 'getLog' gets the patch name and long description from the commandline
-- flags, an existing log file, a pipe, and/or a text editor.
-- It ensures the patch name is not empty nor starts with the prefix TAG.
getLog :: forall prim wX wY . (Patchy prim, PrimPatch prim)
       => [DarcsFlag]                            -- ^ commandline options
       -> Maybe (String, [String])               -- ^ possibly an existing patch name and long description
       -> IO String                              -- ^ temporary log file path
       -> FL prim wX wY                          -- ^ changes to record
       -> IO (String, [String], Maybe String)    -- ^ patch name, long description and possibly the path
                                                 --   to the temporary file that should be removed later
getLog opts m_old make_log chs = gl opts
    where patchname_specified = patchname_helper opts
          patchname_helper (PatchName n:_) | badName n = NoPatchName
                                           | otherwise = FlagPatchName n
          patchname_helper (_:fs) = patchname_helper fs
          patchname_helper [] = case m_old of Just (p,_) -> PriorPatchName p
                                              Nothing    -> NoPatchName
          badName "" = True
          badName n  = "TAG" `isPrefixOf` n
          default_log = case m_old of
                          Nothing    -> []
                          Just (_,l) -> l
          gl (Pipe:_) = do p <- case patchname_specified of
                                  FlagPatchName p  -> return p
                                  PriorPatchName p -> return p
                                  NoPatchName      -> prompt_patchname False
                           putStrLn "What is the log?"
                           thelog <- lines `fmap` Ratified.hGetContents stdin
                           return (p, thelog, Nothing)
          gl (LogFile f:fs) =
              do -- round 1 (patchname)
                 mlp <- lines `fmap` readLocaleFile f `catch` (\(_ :: IOException) -> return [])
                 firstname <- case (patchname_specified, mlp) of
                                (FlagPatchName  p, []) -> return p
                                (_, p:_)               -> if badName p
                                                            then prompt_patchname True
                                                            else return p -- logfile trumps prior!
                                (PriorPatchName p, []) -> return p
                                (NoPatchName, [])      -> prompt_patchname True
                 -- round 2
                 append_info f firstname
                 when (EditLongComment `elem` fs) (void $ editFile f)
                 (name, thelog, _) <- read_long_comment f firstname
                 let toRemove = if willRemoveLogFile opts
                        then Just $ toFilePath f
                        else Nothing
                 return (name, thelog, toRemove)
          gl (EditLongComment:_) =
                  case patchname_specified of
                    FlagPatchName  p -> actually_get_log p
                    PriorPatchName p -> actually_get_log p
                    NoPatchName      -> actually_get_log ""
          gl (NoEditLongComment:_) =
                  case patchname_specified of
                    FlagPatchName  p -> return (p, default_log, Nothing) -- record (or amend) -m
                    PriorPatchName p -> return (p, default_log, Nothing) -- amend
                    NoPatchName      -> do p <- prompt_patchname True -- record
                                           return (p, [], Nothing)
          gl (PromptLongComment:_) =
                  case patchname_specified of
                    FlagPatchName p -> prompt_long_comment p -- record (or amend) -m
                    PriorPatchName p -> prompt_long_comment p
                    NoPatchName -> prompt_patchname True >>= prompt_long_comment
          gl (_:fs) = gl fs
          gl [] = case patchname_specified of
                    FlagPatchName  p -> return (p, default_log, Nothing)  -- record (or amend) -m
                    PriorPatchName "" -> actually_get_log ""
                    PriorPatchName p -> return (p, default_log, Nothing)
                    NoPatchName -> actually_get_log ""
          prompt_patchname retry =
            do n <- askUser "What is the patch name? "
               if badName n
                  then if retry then prompt_patchname retry
                                else fail "Bad patch name!"
                  else return n
          prompt_long_comment oldname =
            do y <- promptYorn "Do you want to add a long comment?"
               if y then actually_get_log oldname
                    else return (oldname, [], Nothing)
          actually_get_log p = do logf <- make_log
                                  -- TODO: make sure encoding used for logf is the same everywhere
                                  -- probably should be locale because the editor will assume it
                                  writeLocaleFile logf $ unlines $ p : default_log
                                  append_info logf p
                                  _ <- editFile logf
                                  l@(name,long,_) <- read_long_comment logf p
                                  if badName name
                                    then do putStrLn "WARNING: empty or incorrect patch name!"
                                            pn <- prompt_patchname True
                                            return (pn, long, Nothing)
                                    else return l
          read_long_comment :: FilePathLike p => p -> String -> IO (String, [String], Maybe p)
          read_long_comment f oldname =
              do f' <- readLocaleFile f
                 let t = filter (not.("#" `isPrefixOf`)) $ (lines.filter (/='\r')) f'
                 case t of []     -> return (oldname, [], Just f)
                           (n:ls) -> return (n, ls, Just f)
          append_info f oldname =
              do fc <- readLocaleFile f
                 appendToFile f $ \h ->
                     do case fc of
                          _ | null (lines fc) -> B.hPut h (encodeLocale (oldname ++ "\n"))
                            | last fc /= '\n' -> B.hPut h (encodeLocale "\n")
                            | otherwise       -> return ()
                        hPutDocLn h
                             $ text "# Please enter the patch name in the first line, and"
                            $$ text "# optionally, a long description in the following lines."
                            $$ text "#"
                            $$ text "# Lines starting with '#' will be ignored."
                            $$ text "#"
                            $$ text "#"
                            $$ text "# This patch contains the following changes:"
                            $$ text "#"
                            $$ prefixLines (text "#") (summaryFL chs)

askAboutDepends :: forall p wR wU wT wY . (RepoPatch p, ApplyState p ~ Tree)
                => Repository p wR wU wT -> FL (PrimOf p) wT wY -> [DarcsFlag]
                -> [PatchInfo] -> IO [PatchInfo]
askAboutDepends repository pa' opts olddeps = do
  -- ideally we'd just default the olddeps to yes but still ask about them.
  -- SelectChanges doesn't currently (17/12/09) offer a way to do this so would
  -- have to have this support added first.
  pps <- readTentativeRepo repository
  pa <- n2pia `fmap` anonymous (fromPrims pa')
  FlippedSeal ps <- return
                      ((case pps of
                          PatchSet x _ -> FlippedSeal (reverseRL x+>+(pa:>:NilFL)))
                         :: FlippedSeal (FL (PatchInfoAnd p)) wY)
  let (pc, tps) = patchChoicesTps ps
      tas = case catMaybes (mapFL (\tp -> if pa `unsafeCompare` tpPatch tp || info (tpPatch tp) `elem` olddeps
                                          then Just (tag tp) else Nothing) tps) of

                [] -> error "askAboutDepends: []"
                tgs -> tgs
  Sealed2 ps' <- return $ case getChoices (forceFirsts tas pc) of _ :> mc :> _ -> Sealed2 $ mapFL_FL tpPatch mc
  (deps:>_) <- runSelection (selectChanges FirstReversed ps') $
                                        selectionContext "depend on" (filter askdep_allowed opts) Nothing Nothing
  return $ olddeps `union` mapFL info deps
 where
       askdep_allowed = not . patchSelectFlag


onlySuccessfulExits :: ExitCode -> Maybe ()
onlySuccessfulExits ExitSuccess = Just ()
onlySuccessfulExits _ = Nothing

recordHelp'' :: String
recordHelp'' =
 "If a test command has been defined with `darcs setpref`, attempting to\n" ++
 "record a patch will cause the test command to be run in a clean copy\n" ++
 "of the working tree (that is, including only recorded changes).  If\n" ++
 "the test fails, you will be offered to abort the record operation.\n" ++
 "\n" ++
 "The `--set-scripts-executable` option causes scripts to be made\n" ++
 "executable in the clean copy of the working tree, prior to running the\n" ++
 "test.  See `darcs get` for an explanation of the script heuristic.\n" ++
 "\n" ++
 "If your test command is tediously slow (e.g. `make all`) and you are\n" ++
 "recording several patches in a row, you may wish to use `--no-test` to\n" ++
 "skip all but the final test.\n" ++
 "\n" ++
 "To see some context (unchanged lines) around each change, use the\n" ++
 "`--unified` option.\n"

